# File: /splang/src/splang/__init__.py
from .interpreter import SplangInterpreter
from .utils import get_last_second, get_first_second, process_ms, get_first_ascii_character
from .opcodes import OPCODES
from .errors import SplangError, SplangWarning, InvalidOpcodeWarning

__version__ = "1.0.8"
__author__ = "Naman Satish"
__all__ = [
    "SplangInterpreter",
    "get_last_second",
    "get_first_second",
    "process_ms",
    "get_first_ascii_character",
    "OPCODES",
    "SplangError",
    "SplangWarning",
    "InvalidOpcodeWarning"
]


import pytest
from splang.utils import get_last_second, get_first_second, process_ms, get_first_ascii_character


def test_get_last_second_valid():
    assert get_last_second("5:32") == 2
    assert get_last_second("10:09") == 9
    with pytest.raises(ValueError):
        get_last_second("invalid")


def test_get_first_second_valid():
    assert get_first_second("5:32") == 3
    assert get_first_second("0:05") == 0
    with pytest.raises(ValueError):
        get_first_second("5")


def test_process_ms_rounding_and_format():
    assert process_ms(0) == "0:00"
    assert process_ms(500) == "0:01"
    assert process_ms(1000) == "0:01"
    assert process_ms(15500) == "0:16"
    assert process_ms(61500) == "1:02"
    assert process_ms(59999) == "1:00"


def test_get_first_ascii_character():
    assert get_first_ascii_character("hello") == "h"
    assert get_first_ascii_character("123!abc") == "1"
    # non-ascii characters should return inverted question mark
    assert get_first_ascii_character("¡¢£") == "¿"
    assert get_first_ascii_character("") == "¿"



import math
def get_last_second(duration_min):
    """
    Extract the last digit of the seconds part of a "mm:ss" string.
    e.g. "5:32" -> 2
    """
    try:
        return int(duration_min.split(':')[1][-1])
    except Exception:
        raise ValueError(f"Invalid duration_min '{duration_min}'")


def get_first_second(duration_min):
    """
    Extract the first digit of the seconds part of a "mm:ss" string.
    e.g. "5:32" -> 3
    """
    try:
        return int(duration_min.split(':')[1][0])
    except Exception:
        raise ValueError(f"Invalid duration_min '{duration_min}'")
    
def process_ms(duration_ms):
    totalSec = math.floor(duration_ms / 1000 + 0.5)
    minutes = totalSec // 60
    seconds = totalSec % 60
    first_seconds = seconds // 10
    last_seconds = seconds % 10
    return f"{minutes}:{first_seconds}{last_seconds}"

def get_first_ascii_character(string):
    for char in string:
        if char.isascii():
            return char
    return "¿"


from .opcodes import OPCODES
from .errors import SplangError, SplangWarning, InvalidOpcodeWarning
import time
import warnings

class SplangInterpreter:
    """
    Interpreter for the Splang (Sound Playlist Esolang) language.

    Accepts a playlist as a list of dictionaries, each containing:
      - track_id (unique identifier)
      - duration_min ("mm:ss" string)
      - opcode (int 00-59)
      - last_second (int 0-9)
      - first_second (int 0-9)
      - first_letter (single character)
      - track_name (string)

    The list order defines the program sequence.
    """

    def __init__(self, playlist):
        self.playlist = playlist
        self.prev_pc = None
        self.pc = 0
        self.stack = []
        self.ra_stack = []
        self.heap = {}
        self.labels = {}
        self.running = True
        #Check if playlist had required fields
        for song in self.playlist:
            if not all(k in song for k in ['track_id', 'duration_min', 'opcode', 'last_second', 'first_second', 'first_letter']):
                raise ValueError(f"Missing required fields in song: {song}")
            
        # Process labels in the playlist
        self.process_labels()

    def run(self):
        """Run until HALT or out-of-bounds."""
        while self.running and 0 <= self.pc < len(self.playlist):
            self.step()
        return self

    
    def step(self):
        #print(f"PC={self.pc} stack={self.stack} ra_stack={self.ra_stack} heap={self.heap} labels={self.labels}")
        row = self.playlist[self.pc]
        cmd, param_count = OPCODES.get(row['opcode'], (None, 0))
        if cmd is None:
            #print(f"PC={self.pc} opcode={row['opcode']} row={row}")
            raise ValueError(f"Unknown opcode {row['opcode']} at PC={self.pc}")

        params = []
        for i in range(1, param_count + 1):
            idx = self.pc + i
            if idx >= len(self.playlist):
                raise IndexError(f"Not enough params at PC={self.pc}")
            params.append(self.playlist[idx])

        method = getattr(self, cmd.lower())
        self.prev_pc = self.pc
        jumped = method(params)
        if jumped is not True:
            self.pc = self.prev_pc + 1 + param_count
        return self._state()

    def _state(self):
        return {
            'pc': self.pc,
            'stack': list(self.stack),
            'ra_stack': list(self.ra_stack),
            'heap': dict(self.heap),
            'labels': dict(self.labels),
            'completed_pc' : self.prev_pc,
            'completed_id' : self.playlist[self.prev_pc]['track_id'] if self.prev_pc < len(self.playlist) else None,
            'completed_op': self.playlist[self.prev_pc]['opcode'] if self.prev_pc < len(self.playlist) else None,
            'next_pc': self.pc if self.pc < len(self.playlist) else None,
            'next_id': self.playlist[self.pc]['track_id'] if self.pc < len(self.playlist) else None,
            'next_op': self.playlist[self.pc]['opcode'] if self.pc < len(self.playlist) else None,
            'running': self.running
        }
    
    def process_labels(self):
        """Process labels in the playlist."""
        for i, song in enumerate(self.playlist):
            if song['opcode'] == 2:
                #Get the next track_id as label
                next_idx = i + 1
                if next_idx < len(self.playlist):
                    label_id = self.playlist[next_idx]['track_id']
                    self.labels[label_id] = next_idx + 1
                else:
                    raise IndexError(f"Label at PC={i} is missing a track_id for jump.")

    def heap_get(self, track_id):
        """Get value from heap by track_id."""
        if track_id in self.heap:
            return self.heap[track_id]
        else:
            raise KeyError(f"Track ID '{track_id}' not found in heap.")
        
    def heap_set(self, track_id, value):
        """Set value in heap by track_id."""
        self.heap[track_id] = value
        
    # === Instruction implementations ===
    def nop(self, params): pass
    def halt(self, params): self.running = False; return True
    def label(self, params): self.labels[params[0]['track_id']] = self.pc + 1 + len(params)
    def jump(self, params): self.pc = self.labels[params[0]['track_id']]; return True
    def jumpz(self, params):
        if self.stack and self.stack[-1] == 0:
            self.pc = self.labels[params[0]['track_id']]
            return True
    def jumpnz(self, params):
        if self.stack and self.stack[-1] != 0:
            self.pc = self.labels[params[0]['track_id']]
            return True
    def jumpz_heap(self, params):
        if self.heap_get(params[0]['track_id']) == 0:
            self.pc = self.labels[params[0]['track_id']]
            return True
    def jumpnz_heap(self, params):
        if self.heap_get(params[0]['track_id']) != 0:
            self.pc = self.labels[params[0]['track_id']]
            return True
    def call(self, params):
        self.ra_stack.append(self.pc + 1 + len(params))
        self.pc = self.labels[params[0]['track_id']]
        return True
    def return_(self, params):
        if self.ra_stack:
            self.pc = self.ra_stack.pop()
            return True
        self.running = False
    def add(self, params): a,b = self.stack.pop(), self.stack.pop(); self.stack.append(b+a)
    def sub(self, params): a,b = self.stack.pop(), self.stack.pop(); self.stack.append(b-a)
    def mul(self, params): a,b = self.stack.pop(), self.stack.pop(); self.stack.append(b*a)
    def div(self, params): a,b = self.stack.pop(), self.stack.pop(); self.stack.append(b//a)
    def mod(self, params): a,b = self.stack.pop(), self.stack.pop(); self.stack.append(b%a)
    def pow(self, params): a,b = self.stack.pop(), self.stack.pop(); self.stack.append(b**a)
    def not_implemented(self, params): warnings.warn(InvalidOpcodeWarning(f"Opcode {self.playlist[self.pc]['opcode']} not implemented @ PC={self.pc}"))
    def push_ls(self, params): self.stack.append(int(params[0]['last_second']))
    def push_fs(self, params): self.stack.append(int(params[0]['first_second']))
    def shift_r_ls(self, params): a=self.stack.pop(); self.stack.append((a>>int(params[0]['last_second'])) + int(params[0]['last_second']))
    def shift_l_ls(self, params): a=self.stack.pop(); self.stack.append((a<<int(params[0]['last_second'])) + int(params[0]['last_second']))
    def shift_r_fs(self, params): a=self.stack.pop(); self.stack.append((a>>int(params[0]['first_second'])) + int(params[0]['first_second']))
    def shift_l_fs(self, params): a=self.stack.pop(); self.stack.append((a<<int(params[0]['first_second'])) + int(params[0]['first_second']))
    def pop(self, params): self.stack.pop()
    def dup(self, params): self.stack.append(self.stack[-1])
    def swap(self, params): a,b=self.stack.pop(),self.stack.pop(); self.stack.extend([a,b])
    def store(self, params): self.heap_set(params[0]['track_id'], self.stack.pop())
    def store_top(self, params): val=self.stack.pop(); addr=self.stack.pop(); self.heap[addr] = val
    def load(self, params): self.stack.append(self.heap_get(params[0]['track_id']))
    def load_top(self, params): addr=self.stack.pop(); self.stack.append(self.heap.get(addr, 0))
    def inc_heap(self, params): self.heap_set(params[0]['track_id'], self.heap_get(params[0]['track_id']) + 1)
    def dec_heap(self, params): self.heap_set(params[0]['track_id'], self.heap_get(params[0]['track_id']) - 1)
    def inc(self, params): self.stack.append(self.stack.pop() + 1)
    def dec(self, params): self.stack.append(self.stack.pop() - 1)
    def stdin_int(self, params): val=int(input()); self.stack.append(val)
    def stdin(self, params): s=input(); [self.stack.append(ord(ch)) for ch in s]
    def stdout_int(self, params): print(self.stack.pop())
    def stdout(self, params): print(chr(self.stack.pop()), end='')
    def read_char(self, params): self.stack.append(ord(params[0]['first_letter']))
    def listen(self, params): time.sleep(int(params[0]['duration_min'].split(':')[0]) * 60 + int(params[0]['duration_min'].split(':')[1]))
    def and_(self, params): a,b=self.stack.pop(),self.stack.pop(); self.stack.append(b&a)
    def or_(self, params): a,b=self.stack.pop(),self.stack.pop(); self.stack.append(b|a)
    def xor(self, params): a,b=self.stack.pop(),self.stack.pop(); self.stack.append(b^a)
    def not_(self, params): a=self.stack.pop(); self.stack.append(~a)
    def equal(self, params): a,b=self.stack.pop(),self.stack.pop(); self.stack.append(1 if b==a else 0)
    def not_equal(self, params): a,b=self.stack.pop(),self.stack.pop(); self.stack.append(1 if b!=a else 0)
    def greater(self, params): a,b=self.stack.pop(),self.stack.pop(); self.stack.append(1 if b>a else 0)
    def less(self, params): a,b=self.stack.pop(),self.stack.pop(); self.stack.append(1 if b<a else 0)
    def greater_equal(self, params): a,b=self.stack.pop(),self.stack.pop(); self.stack.append(1 if b>=a else 0)
    def less_equal(self, params): a,b=self.stack.pop(),self.stack.pop(); self.stack.append(1 if b<=a else 0)


import pytest
import warnings
from splang.interpreter import SplangInterpreter
from splang.errors import SplangError, SplangWarning, InvalidOpcodeWarning

@ pytest.fixture
def song_factory():
     def make_song(track_id='~', duration_min="0:00", first_letter='~'):
        """
            Create a testing song with the required fields:
            - track_id (string)
            - opcode (int 0-59)
            - last_second (int 0-9)
            - first_second (int 0-5)
            - first_letter (single character)
            - duration_min (string in format "mm:ss")
        """
        if not isinstance(track_id, str):
            raise TypeError("track_id must be a string")
        if not isinstance(first_letter, str) or len(first_letter) != 1:
            raise ValueError("first_letter must be a single character string")
        #Check split duration_min into minutes and seconds
        if not isinstance(duration_min, str) or not (
            len(duration_min.split(':')) == 2 and
            all(part.isdigit() for part in duration_min.split(':')) and
            0 <= int(duration_min.split(':')[1]) < 60
        ):
            raise ValueError("duration_min must be a string in format 'mm:ss' with valid minutes and seconds")
        last_second = int(duration_min.split(':')[1]) % 10
        first_second = int(duration_min.split(':')[1]) // 10
        opcode = 10 * first_second + last_second
        return {
            'track_id': track_id,
            'opcode': opcode,
            'last_second': last_second,
            'first_second': first_second,
            'first_letter': first_letter,
            'duration_min': duration_min
        }
     return make_song


def test_nop_and_halt(song_factory):
    # NOP then HALT
    p = [ song_factory(duration_min="0:00"), # NOP
          song_factory(duration_min="0:01"), # HALT
          song_factory('s1', "0:20"),  # PUSH_LS
          song_factory('p1', "0:04"),  # param1 with last_second 4
        ]
    #print(p)
    interp = SplangInterpreter(p)
    interp.run()
    assert not interp.running
    assert interp.pc == 1  # Should stop at HALT
    assert interp.stack == []  # Stack should be empty after HALT


def test_push_ls_and_add(song_factory):
    p = []
    p.append(song_factory('s0', "0:20")) # PUSH_LS
    p.append(song_factory('p1', "0:02"))  # param1 with last_second 2
    p.append(song_factory('s1', "0:20")) # PUSH_LS
    p.append(song_factory('p2', "0:03"))  # param2 with last_second 3
    p.append(song_factory('s2', "0:10")) # ADD
    p.append(song_factory('s3', "0:01")) # HALT
    interp = SplangInterpreter(p)
    interp.run()
    assert interp.stack == [5]



def test_store_and_load_heap(song_factory):
    p = []
    key = 'keytrack'
    p.append(song_factory('s0', "0:20"))  # PUSH_LS
    p.append(song_factory('p1', "0:04"))  # param1 with last_second 4
    p.append(song_factory('s1', "0:30"))  # STORE
    p.append(song_factory(key, "0:00"))  # key track
    p.append(song_factory('s2', "0:32"))  # LOAD
    p.append(song_factory(key, "0:00"))  # key track
    p.append(song_factory('s3', "0:01"))  # HALT
    interp = SplangInterpreter(p)
    interp.run()
    assert interp.stack == [4]


def test_undef_opcode_raises(song_factory):
    """
    Test that an undefined opcode raises a ValueError.
    """
    p = [
         song_factory('s1', "0:16"),  # Undefined opcode
    ]
    interp = SplangInterpreter(p)
    # assert runs without raising an error
    with pytest.warns(InvalidOpcodeWarning):
        interp.run()

def test_unknown_opcode_raises(song_factory):
    """
    Test that an unknown opcode raises a ValueError.
    """
    p = [
         song_factory('s1', "0:01"),  # Defined opcode
    ]
    p[0]['opcode'] = 99  # Unknown opcode
    interp = SplangInterpreter(p)
    with pytest.raises(ValueError):
        interp.run()

def test_unknown_heap_key_raises(song_factory):
    """
    Test that trying to load a non-existent heap key raises a KeyError.
    """
    p = [
            song_factory('s1', "0:20"),  # PUSH_LS
            song_factory('p1', "0:04"),  # param1 with last_second 4
            song_factory('s2', "0:32"),  # LOAD
            song_factory('nonexistent_key', "0:00"),  # Non-existent key
            song_factory('s3', "0:01")   # HALT
        ]
    interp = SplangInterpreter(p)
    with pytest.raises(KeyError):
        interp.run()

def test_jump_to_label(song_factory):
    """
    Test that jumping to a label works correctly.
    """
    p = [
        song_factory('s1', "0:03"),  # JUMP
        song_factory('label1', "0:00"),  # param1 with track_id = 'label1'
        song_factory('s2', "0:01"),  # HALT
        song_factory('s3', "0:02"),  # LABEL
        song_factory('label1', "0:00"),  # param1 with track_id = 'label1'
        song_factory('s4', "0:20"),  # PUSH_LS
        song_factory('p1', "0:04"),  # param1 with last_second 4
        song_factory('s5', "0:01")   # HALT
    ]
    interp = SplangInterpreter(p)
    interp.run()
    assert interp.stack == [4]


# splang/src/splang/cli.py
import json
import sys
import spotipy
import os
import importlib.resources
from splang.utils import get_last_second, get_first_second, process_ms, get_first_ascii_character
from splang.interpreter import SplangInterpreter

def get_all_playlists(sp):
    playlists = []
    results = sp.current_user_playlists()
    while results:
        playlists.extend(results['items'])
        if results['next']:
            results = sp.next(results)
        else:
            break
    return playlists

def get_all_tracks_from_playlist(sp, playlist_id):
    all_tracks = []
    results = sp.playlist_tracks(playlist_id)
    while results:
        all_tracks.extend(results['items'])
        if results['next']:
            results = sp.next(results)
        else:
            break
    return all_tracks

# Get all saved tracks
def get_all_saved_tracks(sp):
    tracks = []
    results = sp.current_user_saved_tracks()
    while results:
        tracks.extend(results['items'])
        if results['next']:
            results = sp.next(results)
        else:
            break
    return tracks

def main():
    # Check if there is the -h flag
    if len(sys.argv) > 1 and sys.argv[1] == '-h':
        print("Usage: splang [path_to_playlist.json]")
        print("Use: splang -hello to use the hello world example.")
        print("If no path is provided, the program will prompt you to log into Spotify and select a playlist.")
        sys.exit(0)


    # If a path is provided, load the playlist from the file
    if len(sys.argv) > 1:
        path = sys.argv[1]
        try:
            if sys.argv[1] == '-hello':
                with importlib.resources.files("splang.examples").joinpath("helloworld.json").open("r") as f:
                    track_data = json.load(f)
                for i, track in enumerate(track_data):
                    print(f"Track {i}: {track['track_name']} - {track['duration_min']}")
            else:
                with open(path, 'r') as f:
                    track_data = json.load(f)
        except FileNotFoundError:
            print(f"Error: File {path} not found.")
            sys.exit(1)
        except json.JSONDecodeError:
            print(f"Error: File {path} is not a valid JSON file.")
            sys.exit(1)

        if not isinstance(track_data, list):
            print(f"Error: File {path} does not contain a valid playlist format. Expected a list of tracks.")
            sys.exit(1)

        #Process track_id, duration_min, first_letter ; Generate opcode, first_second, last_second
        for i, track in enumerate(track_data):
            if not isinstance(track, dict):
                print(f"Error: Track {i} in the playlist is not a valid dictionary.")
                sys.exit(1)

            # If track_id is not valid, set it to a default value
            if 'track_id' not in track or not isinstance(track['track_id'], str):
                track_data[i]['track_id'] = f'track_{i}'
            
            # If first_letter is not valid, set it to a default value
            if 'first_letter' not in track or not isinstance(track['first_letter'], str) or len(track['first_letter']) != 1:
                track_data[i]['first_letter'] = '¿'

            # If duration_min is not valid, set it to a default value
            if 'duration_min' not in track or not isinstance(track['duration_min'], str):
                track_data[i]['duration_min'] = "0:00"

            duration_min = track['duration_min']
            if not isinstance(duration_min, str) or not (len(duration_min.split(':')) == 2 and
                                                         all(part.isdigit() for part in duration_min.split(':')) and
                                                         0 <= int(duration_min.split(':')[1]) < 60):
                print(f"Error: duration_min must be a string in format 'mm:ss' with valid minutes and seconds in track: {track}")
                sys.exit(1)

            # Calculate opcode, first_second, last_second
            last_second = int(duration_min.split(':')[1]) % 10
            first_second = int(duration_min.split(':')[1]) // 10
            opcode = 10 * first_second + last_second
            track['opcode'] = opcode
            track['first_second'] = first_second
            track['last_second'] = last_second

        
    else:
        #Prompt user to log into Spotify
        #Check if SPOTIPY_CLIENT_ID, SPOTIPY_CLIENT_SECRET, and SPOTIPY_REDIRECT_URI are set
        if 'SPOTIPY_CLIENT_ID' not in os.environ or 'SPOTIPY_CLIENT_SECRET' not in os.environ or 'SPOTIPY_REDIRECT_URI' not in os.environ:
            print("Error: SPOTIPY_CLIENT_ID, SPOTIPY_CLIENT_SECRET, and SPOTIPY_REDIRECT_URI must be set in environment variables.")
            print("Please set them and try again.")
            sys.exit(1)
        scope = "user-library-read playlist-read-private playlist-read-collaborative"
        sp = spotipy.Spotify(auth_manager=spotipy.SpotifyOAuth(os.environ['SPOTIPY_CLIENT_ID'],
                                                            os.environ['SPOTIPY_CLIENT_SECRET'],
                                                            os.environ['SPOTIPY_REDIRECT_URI'],
                                                            scope=scope))
        if not sp:
            print("Error: Could not authenticate with Spotify. Please check your credentials.")
            sys.exit(1)

        # Check if the user is logged in
        if not sp.current_user():
            print("Error: You are not logged in to Spotify. Please log in and try again.")
            sys.exit(1)


        # Check if the user has any playlists

        all_tracks = []
        choices = get_all_playlists(sp)
        if len(choices) != 0:
            playlists = []
            # Prompt the user to type which 1 playlist they want to use
            print("Please select the playlist you want to use:")
            for i, playlist in enumerate(choices):
                print(f"{i}: {playlist['name']}")
                playlists.append(playlist)

            # Prompt the user to select one playlist from the options
            selected_playlist = input("Please select a playlist by number: ").strip().lower()
            if selected_playlist.isdigit() and int(selected_playlist) < len(playlists):
                selected_playlist = int(selected_playlist)
                all_tracks = get_all_tracks_from_playlist(sp, playlists[selected_playlist]['id'])
            else:
                print("Invalid selection. Please try again.")
                sys.exit(1)
        else:
            print("No playlists found. Please create a playlist and try again.")
            sys.exit(1)

        track_data = []
        for item in all_tracks:
            track = item['track']
            track_name = track['name']
            track_id = track['id']
            artist_name = track['artists'][0]['name']
            album_name = track['album']['name']
            release_date = track['album']['release_date']
            try:
                spotify_url = track['external_urls']['spotify']
            except KeyError:
                # Handle the case where the Spotify URL is not available
                spotify_url = None

            try:
                # Handle the case where the track does not have a popularity score
                popularity = track['popularity']
            except KeyError:
                # Set a default value for popularity if not available
                popularity = None
            duration_ms= track['duration_ms']
            duration_min = process_ms(duration_ms)
            first_second = get_first_second(duration_min)
            last_second = get_last_second(duration_min)
            first_letter = get_first_ascii_character(track_name)
            track_data.append({
                'track_name': track_name,
                'track_id': track_id,
                'artist_name': artist_name,
                'album_name': album_name,
                'release_date': release_date,
                'spotify_url': spotify_url,
                'duration_ms': duration_ms,
                'popularity': popularity,
                'duration_min': duration_min,
                'first_letter': first_letter,
                'opcode': 10 * first_second + last_second,
                'first_second': first_second,
                'last_second': last_second
            })
        
        # Ask the user if they want to save the playlist to a JSON file
        save_to_json = input("Do you want to save the playlist to a JSON file? (y/n): ").strip().lower()
        if save_to_json != 'y':
            print("Playlist not saved to JSON file.")
            # Continue without saving
            pass
        else:
            # Save to cwd
            json.dump(track_data, open(os.path.join(os.getcwd(), playlists[selected_playlist]['name'] + '.json'), 'w'), indent=4)
            print("Playlist saved to JSON file.")

    # Run the Splang interpreter
    try:
        interp = SplangInterpreter(track_data)
    except Exception as e:
        print("Error initializing the interpreter:", e, file=sys.stderr)
        sys.exit(1)

    try:
        interp.run()
        print("\n")
    except Exception as e:
        print("Error during execution:", e, file=sys.stderr)
        print("Interpreter state at the time of error:")
        print(interp._state())
        sys.exit(1)


        


class SplangError(Exception):
    """Base class for Splang interpreter errors."""
    def __init__(self, message):
        super().__init__(message)
        self.message = message

class SplangWarning(Warning):
    """Base class for warnings in the Splang interpreter."""
    def __init__(self, message):
        super().__init__(message)
        self.message = message

class InvalidOpcodeWarning(Warning):
    """Warning for invalid opcodes in the Splang interpreter."""
    def __init__(self, opcode):
        super().__init__(f"Invalid opcode: {opcode}")
        self.opcode = opcode